import json
from pathlib import Path
from typing import Any, MutableMapping


def json_loader(filepath: Path) -> MutableMapping[str, Any]:
    """
    Open and load a config mapping from a JSON file.

    :param filepath: The path of the file to load from.
    :return: The mapping of keys to value loaded from the file.
    """
    with filepath.open("r") as f:
        return json.load(f)



from pathlib import Path
from typing import Any, MutableMapping

import toml


def toml_loader(filepath: Path) -> MutableMapping[str, Any]:
    """Open and load a dict from a TOML file."""
    with filepath.open("r") as f:
        return toml.load(f)



from os import environ
from pathlib import Path
from typing import MutableMapping

from dotenv import load_dotenv


def env_loader(filepath: Path) -> MutableMapping[str, str]:
    """
    Load the given *.env file and the user's environment variables.

    Prioritizes system/user environment variables. Will not override any existing environment variables from the
    system.

    :param filepath: The path of the file to load from.
    :return: The mapping of keys to value loaded from the env file and user's environment.
    """
    load_dotenv(filepath)

    return environ



from typing import Any, MutableMapping, Optional


def merge_dicts(a: MutableMapping[str, Any], b: MutableMapping[str, Any], path: Optional[list] = None) -> MutableMapping[str, Any]:
    """
    Merge the keys and values of the two dicts.

    :param a:
    :param b:
    :param path:
    :return:
    :raises ValueError: When both dicts assign the same key, with different values.
    """
    if path is None:
        path = []

    for key in b:

        if key not in a:
            a[key] = b[key]
            continue

        if isinstance(a[key], dict) and isinstance(b[key], dict):
            merge_dicts(a[key], b[key], path + [str(key)])
        elif a[key] == b[key]:
            pass  # same leaf value
        else:
            raise ValueError(f"Conflict at {'.'.join(path + [str(key)])}")

    return a



from logging import getLogger
from pathlib import Path
from typing import Any, Dict, List, MutableMapping, Optional, Type, TypeVar, Union

from pydantic import BaseModel
from typing_inspect import get_origin

from ._util import merge_dicts
from .loader import Loader, auto_loader

logger = getLogger(__name__)

_config: MutableMapping[str, Any] = {}


def register_config(
    filepath: Union[str, Path],
    prefix: Optional[str] = None,
    loader: Loader = auto_loader,
    strict: bool = False
) -> None:
    """
    Register a config file to load from.

    :param filepath: Path to the config to load.
    :param prefix: Prefix all config values loaded.
    :param loader: Loader to user for gathering values from config file.
        The default `auto_loader` will resolve a loader from the config file's extension.
    :param strict: If strict, any duplicate keys will raise a ValueError.
        If not strict, duplicate keys will attempt to be merged if both values are dicts.
    :return:
    """
    if isinstance(filepath, str):
        filepath = Path(filepath)

    if not filepath.match("*.env") and not filepath.exists():
        raise FileNotFoundError(filepath)

    raw_config = loader(filepath)

    if prefix:
        raw_config = {prefix: raw_config}

    for key, val in raw_config.items():
        current_value = _config.get(key)

        if current_value is None:
            _config[key] = val
            continue

        if strict:
            raise ValueError("Conflicting config")

        logger.warning("Conflicting config value, merging")

        if not isinstance(current_value, dict):
            logger.warning("Dropping conflicting config value. [%s:%s]", key, val)
            continue

        _config[key] = merge_dicts(current_value, val)


def get_raw_config() -> MutableMapping[str, Any]:
    """:return: The current loaded config."""
    return _config.copy()


T = TypeVar("T")


class _ResolveFromConfig(type):
    def __call__(cls, variable_path: str, _type: Type[T], deliminator: str = ".") -> T:

        config_value_path = variable_path.split(deliminator)

        data = _config

        for step in config_value_path:
            try:
                data = data[step]
            except KeyError:
                raise KeyError(f"Config: {variable_path} does not exist")

        if get_origin(_type) in (List, Dict):
            return _type(data)
        if issubclass(_type, BaseModel):
            return _type(**data)
        else:
            return _type(data)


class ConfigValue(metaclass=_ResolveFromConfig):
    """Resolve to a value defined within the config."""
